import logging
import psutil
import os
import json
import re
import subprocess
import socket
import atexit
import tempfile
import fnmatch
import requests
import time
import base64
import hashlib
import pathlib
from datetime import datetime
from typing import Optional

from ruxit.api.exceptions import AuthException, ConfigException
from ruxit.api.base_plugin import RemoteBasePlugin

logger = logging.getLogger(__name__)
MAX_ITEMS = 1000
MAX_EVENTS = 99

logging.basicConfig(level=logging.DEBUG)
logger.setLevel(logging.DEBUG)

def get_java_process(process_filter: list) -> Optional[psutil.Process]:
    for process in psutil.process_iter():
        try:
            #logger.debug(f"Matchin process {process_filter} with process " + "".join(process.cmdline()))
            if all(x in "".join(process.cmdline()) for x in process_filter ):
                logger.debug(f"Found {process_filter} in " + "".join(process.cmdline()))
                return process
        except (OSError, psutil.PermissionError, psutil.AccessDenied):
            continue
    return None


def find_java_executable(java_path):
    locations = [
        f"../../../gateway/jre/bin/java{'.exe' if os.name == 'nt' else ''}",
        f"../../../jre/bin/java{'.exe' if os.name == 'nt' else ''}",
    ]
    
    if java_path != '':
        locations.insert(0, java_path)

    for location in locations:
        if os.path.isfile(location):
            return location

    if os.environ.get("GATEWAY_HOME", False):
        return f"{os.environ.get('GATEWAY_HOME')}/jre/bin/java{'.exe' if os.name == 'nt' else ''}"

    return "java"


def start_java_process(endpoint_id,tempdir, java_path):
    java_executable = find_java_executable(java_path)
    logger.info(f"Using java executable: '{java_executable}'")
    logfile_name = re.sub('[^0-9a-zA-Z]+', '_', endpoint_id)
    current_file_path = os.path.dirname(os.path.realpath(__file__))
    cmd = [
        java_executable,
        f'-Ddynatrace.ibmmq.endpoint="{endpoint_id}"',
        f'-Dlogfile.name={logfile_name}',
        f'-Djavax.net.debug=ssl:handshake',
        "-Xmx400M",
        "-jar",
        os.path.join(current_file_path, "lib", "dynatrace_ibm_mq_connector.jar"),
        f'"{tempdir}"',
        f'"{endpoint_id}"'
    ] 

    logger.info(f"Attempting to execute command: '{' '.join(cmd)}'")

    if os.name == "nt":
        pid = subprocess.Popen(cmd, creationflags=0x00000008).pid
    else:
        pid = subprocess.Popen(cmd, close_fds=True).pid

    logger.debug(f'Process pid = {pid}')
    if psutil.pid_exists(pid):
        for process in psutil.process_iter():
            try:
                if process.pid == pid:
                    logger.debug(f'Found process id {pid}, returning it')
                    return process
            except (psutil.PermissionError, psutil.AccessDenied):
                continue


def stop_java_process(process: psutil.Process):
    logger.info(f"Attempting to terminate process {process}")
    process.terminate()
    gone, alive = psutil.wait_procs([process], timeout=3)
    if alive:
        try:
            process.kill()
        except psutil.NoSuchProcess:
            pass
    gone, alive = psutil.wait_procs([process], timeout=3)
    if alive:
        logger.error(f"Could not kill the process {process}. The plugin state is invalid until the process is terminated")
    else:
        logger.info(f"Process '{process}' terminated")

class EncodeText(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (bytes, bytearray)):
            return obj.decode('ASCII')
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)
        
class IBM_MQ_Java_Plugin(RemoteBasePlugin):
    # Store metrics that need their positive/negative delta calculated
    metricHolder = {}
    OS = {3: 'AIX/Unix', 4: 'IBM i Series', 5: 'Microsoft Windows', 11: 'Microsoft Windows NT', 1: 'z/OS', 13: 'HP Integrity NonStop Server',
         12: 'OpenVMS', 2: 'OS/2', 23: 'z/TPF', 27: 'z/VSE', 28: 'MQ Appliance'}
         
    channel_states = {
                        'INACTIVE': 'Inactive',
                        'MQCHS_INACTIVE': 'Inactive',
                        'MQCHS_BINDING': 'Binding',
                        'MQCHS_STARTING': 'Starting',
                        'MQCHS_RUNNING': 'Running',
                        'MQCHS_PAUSED': 'Paused',
                        'MQCHS_RETRYING': 'Retrying',
                        'MQCHS_STOPPED': 'Stopped',
                        'MQCHS_STOPPING': 'Stopping',
                        'MQCHS_REQUESTING': 'Requesting',
                        'MQCHS_SWITCHING': 'Switching',
                        'MQCHS_INITIALIZING': 'Initializing'}
         
    last_topology_retrieve = {}
    previousEvents = []
    events = 0
    
    def initialize(self, **kwargs):
        self.tempdir = None
        try:
            with open('../../agent/conf/remotepluginlauncher.ini') as fp:
                for line in fp:
                    if '-log.dir' in line or '-logdir' in line:
                        self.tempdir = os.path.join(line.split('=')[1].strip(), kwargs['json_config']['name'])
                        self.tempdir = pathlib.Path(self.tempdir).as_posix()
                        break
        except Exception as e:
            logger.error(f'Unable to read log directory from remotepluginlauncher.ini. Using default tempdir. {e}')
        
        if not self.tempdir:
            import tempfile
            self.tempdir = tempfile.gettempdir()
            
        endpoint_id = f"{kwargs.get('endpoint', 'default_endpoint')}_{kwargs.get('associated_entity', 'default_endpoint')}"
        
        java_process = get_java_process([endpoint_id, "dynatrace_ibm_mq_connector.jar"])
        if java_process is not None:
            stop_java_process(java_process)
        
    
    def query(self, **kwargs) -> None:
        endpoint = kwargs['activation_context'].endpoint_name
        endpoint = re.sub('[^0-9a-zA-Z]+', '_', endpoint)
        config = kwargs['config']
        group = config['group'].strip() if config['group'].strip() != '' else 'Default group'
        hosts = config['host_port'].split(',')
        runResetStats = config['reset_q_stats']
        endpoint_id = f"{endpoint}_{kwargs.get('associated_entity', 'default_endpoint')}"
        channel_totals = {}
        tenant = config['tenant']
        api_token = config['api_token']
        self.queue_manager = config['queue_manager']
        proxy_host = config['http_proxy_host'].strip()
        proxy_user = config['http_proxy_user'].strip()
        proxy_password = config['http_proxy_password']
        dlq_alert = True if 'dlq_alert' in config and str(config['dlq_alert']).lower() == 'true' else False
        retry_channel_alert = True if 'retry_channel_alert' in config and str(config['retry_channel_alert']).lower() == 'true' else False
        self.request_proxy_list = {}
        listeners = []
        channels = []
        queues = []
        
        if config['get_queue_topology'] and (config['tenant'].strip() == '' or config['api_token'].strip() == ''):
            raise ConfigException('You have selected to get queue manager topology but did not provide the cluster/tenant or the API token')
            
        if 'queues' in config and config['queues'] != '':
            queues = config['queues'].split(',',MAX_ITEMS)
            if len(queues) == MAX_ITEMS+1:
                queues.pop()
                
            #Move queues with wildcards to the end to guarantee specific ones first
            moveToStart = filter(lambda x: '*' not in x, queues)
            for item in moveToStart:
                queues.remove(item)
                queues.insert(0, item)
            
            queues = [x.strip() for x in queues if x is not '']
            
        if 'channel' in config and config['channel'] != '':
            channels = config['channel'].split(',',MAX_ITEMS)
            if len(queues) == MAX_ITEMS+1:
                channels.pop()
                
            #Move channels with wildcards to the end to guarantee specific ones first
            moveToStart = filter(lambda x: '*' not in x, channels)
            for item in moveToStart:
                channels.remove(item)
                channels.insert(0, item)
            channels = [x.strip() for x in channels if x is not '']
            
        if 'listener' in config and config['listener'] != '':
            listeners = config['listener'].split(',')
            listeners = [x.strip() for x in listeners if x is not '']
        
        '''
        if 'host_port' in config and config['host_port'] != '':
            hosts = config['host_port'].split(',')
        '''
        
        if self.queue_manager not in self.last_topology_retrieve:
            self.last_topology_retrieve[self.queue_manager] = {'time': 0}
        
        g1 = self.topology_builder.create_group(f'IBM MQ - {group}', f'IBM MQ - {group}')
        self.node = g1.create_element(f'IBM MQ - {group} - {self.queue_manager}', f'IBM MQ - {group} - {self.queue_manager}')
        self.node.report_property("Plugin version", kwargs["json_config"]["version"])
        
        if proxy_host is not '':
            if proxy_user is not '':
                self.request_proxy_list = {'http': f'http://{proxy_user}:{proxy_password}@{proxy_host}/', 'https': f'http://{proxy_user}:{proxy_password}@{proxy_host}/'}
            else:
                self.request_proxy_list = {'http': f'http://{proxy_host}/', 'https': f'http://{proxy_host}/'}

        # Get the host IP address
        hostList = []
        total_hosts = len(hosts)-1
        for i, host in enumerate(hosts):
            matcher = re.search('(.+)\((\d+)\)|(.+):(\d+)',host.strip(),re.M|re.I)
            try:
                hostIP = matcher.group(1) or matcher.group(3)
                hostPort = int(matcher.group(2) or matcher.group(4))
                
                logger.info(f'Host: {hostIP}. Port: {hostPort}')
                
                if i == total_hosts:
                    hostList.append(f'{hostIP}')
                else:
                    hostList.append(f'{hostIP}({hostPort})')
                    
                try:
                    hostIP = socket.gethostbyname(hostIP)
                    self.node.add_endpoint(hostIP, hostPort)
                except socket.error as e:
                    logger.error(f'Unable to resolve IP for host: {host}. Will not add endpoint')
            except Exception as e2:
                raise ConfigException(f'Hostname or IP not valid: {host}')
                logger.exception(e2)

        
        # Set all metrics to retrieve
        metrics = {}
        metrics['qmStatus'] = {'name': 'QueueManager.Status', 'attr': 'MQIACF_Q_MGR_STATUS', 'owner': 'QueueManager', 'metric': 'absolute', 'normalValue' : 'MQSVC_STATUS_RUNNING'}
        metrics['qmConnections'] = {'name': 'QueueManager.Connections', 'json_field': 'connections', 'attr': 'MQIACF_CONNECTION_COUNT', 'owner': 'QueueManager', 'metric': 'absolute'}
        metrics['qmActiveChannels'] = {'name': 'QueueManager.ActiveChannels', 'json_field': 'activeChannels', 'attr': 'MQIACF_CONNECTION_COUNT', 'owner': 'QueueManager', 'metric': 'absolute'}
        
        metrics['queueCurrentQueueDepth'] = {'name': 'Queue.Depth', 'json_field': 'currentQueueDepth', 'paramIdentifier' : 'MQIACF_Q_ATTRS', 'attr': 'MQIA_CURRENT_Q_DEPTH', 'owner': 'Queue', 'metric': 'absolute', 'splitName': 'Queue', 'splitBy': 'MQCA_Q_NAME'}
        metrics['queueInhibitGet'] = {'name': 'Queue.InhibitGet', 'alert': True, 'json_field': 'inhibitGet', 'paramIdentifier' : 'MQIACF_Q_ATTRS', 'attr': 'MQIA_INHIBIT_GET', 'owner': 'Queue', 'metric': 'absolute', 'splitName': 'Queue', 'splitBy': 'MQCA_Q_NAME'}
        metrics['queueInhibitPut'] = {'name': 'Queue.InhibitPut', 'alert': True, 'json_field': 'inhibitPut', 'paramIdentifier' : 'MQIACF_Q_ATTRS', 'attr': 'MQIA_INHIBIT_PUT', 'owner': 'Queue', 'metric': 'absolute', 'splitName': 'Queue', 'splitBy': 'MQCA_Q_NAME'}
        metrics['queueMaxQueueDepth'] = {'json_field': 'maxQueueDepth', 'paramIdentifier' : 'MQIACF_Q_ATTRS', 'attr': 'MQIA_MAX_Q_DEPTH', 'owner': 'Queue', 'metric': 'absolute', 'splitName': 'Queue', 'splitBy': 'MQCA_Q_NAME'}
        metrics['queueOpenInputHandles'] = {'name': 'Queue.OpenInputCount', 'json_field': 'openInputHandles', 'paramIdentifier' : 'MQIACF_Q_ATTRS', 'attr': 'MQIA_OPEN_INPUT_COUNT', 'owner': 'Queue', 'metric': 'absolute', 'splitName': 'Queue', 'splitBy': 'MQCA_Q_NAME'}
        metrics['queueOpenOutputHandles'] = {'name': 'Queue.OpenOutputCount', 'json_field': 'openOutputHandles', 'paramIdentifier' : 'MQIACF_Q_ATTRS', 'attr': 'MQIA_OPEN_OUTPUT_COUNT', 'owner': 'Queue', 'metric': 'absolute', 'splitName': 'Queue', 'splitBy': 'MQCA_Q_NAME'}
        metrics['queueOldestMessageAge'] = {'name': 'Queue.OldestMessageAge', 'json_field': 'oldestMessageAge', 'paramIdentifier' : 'MQIACF_Q_STATUS_ATTRS', 'attr': 'MQIACF_OLDEST_MSG_AGE', 'owner': 'Queue', 'metric': 'absolute', 'unavailable': 'MQMON_NOT_AVAILABLE', 'splitName': 'Queue', 'splitBy': 'MQCA_Q_NAME'}
        metrics['queueUncommittedMessages'] = {'name': 'Queue.UncommittedMessages', 'json_field': 'uncommittedMessages', 'paramIdentifier' : 'MQIACF_Q_STATUS_ATTRS', 'attr': 'MQIACF_UNCOMMITTED_MSGS', 'owner': 'Queue', 'metric': 'absolute', 'splitName': 'Queue', 'splitBy': 'MQCA_Q_NAME'}
        metrics['queueLastGetDate'] = {'name': 'Queue.LastGetDate', 'json_field': 'lastGetDate', 'paramIdentifier' : 'MQIACF_Q_STATUS_ATTRS', 'attr': 'MQCACF_LAST_GET_DATE', 'owner': 'Queue', 'metric': 'absolute', 'splitName': 'Queue', 'splitBy': 'MQCA_Q_NAME'}
        metrics['queueLastGetTime'] = {'name': 'Queue.LastGetTime', 'json_field': 'lastGetTime', 'paramIdentifier' : 'MQIACF_Q_STATUS_ATTRS', 'attr': 'MQCACF_LAST_GET_TIME', 'owner': 'Queue', 'metric': 'absolute', 'splitName': 'Queue', 'splitBy': 'MQCA_Q_NAME'}
        metrics['queueLastPutDate'] = {'name': 'Queue.LastPutDate', 'json_field': 'lastPutDate', 'paramIdentifier' : 'MQIACF_Q_STATUS_ATTRS', 'attr': 'MQCACF_LAST_PUT_DATE', 'owner': 'Queue', 'metric': 'absolute', 'splitName': 'Queue', 'splitBy': 'MQCA_Q_NAME'}
        metrics['queueLastPutTime'] = {'name': 'Queue.LastPutTime', 'json_field': 'lastPutTime', 'paramIdentifier' : 'MQIACF_Q_STATUS_ATTRS', 'attr': 'MQCACF_LAST_PUT_TIME', 'owner': 'Queue', 'metric': 'absolute', 'splitName': 'Queue', 'splitBy': 'MQCA_Q_NAME'}
        if runResetStats == True:
            metrics['queueEnqueue'] = {'name': 'Queue.Enqueue', 'json_field': 'enqueueRate', 'attr': 'MQIA_MSG_DEQ_COUNT', 'owner': 'Queue', 'metric': 'absolute', 'splitName': 'Queue', 'splitBy': 'MQCA_Q_NAME'}
            metrics['queueDequeue'] = {'name': 'Queue.Dequeue', 'json_field': 'dequeueRate', 'attr': 'MQIA_MSG_ENQ_COUNT', 'owner': 'Queue', 'metric': 'absolute', 'splitName': 'Queue', 'splitBy': 'MQCA_Q_NAME'}
        metrics['queueShortQueueTimeIndicator'] = {'name': 'Queue.TimeIndicatorShort', 'json_field': 'shortQueueTimeIndicator', 'paramIdentifier' : 'MQIACF_Q_STATUS_ATTRS', 'attr': 'MQIACF_Q_TIME_INDICATOR', 'owner': 'Queue', 'metric': 'absolute', 'unavailable': 'MQMON_NOT_AVAILABLE', 'splitName': 'Queue', 'splitBy': 'MQCA_Q_NAME'}
        metrics['queueLongQueueTimeIndicator'] = {'name': 'Queue.TimeIndicatorLong', 'json_field': 'longQueueTimeIndicator', 'paramIdentifier' : 'MQIACF_Q_STATUS_ATTRS', 'attr': 'MQIACF_Q_TIME_INDICATOR', 'owner': 'Queue', 'metric': 'absolute', 'unavailable': 'MQMON_NOT_AVAILABLE', 'splitName': 'Queue', 'splitBy': 'MQCA_Q_NAME'}
        
        metrics['channelStatus'] = {'json_field': 'status', 'paramIdentifier' : 'MQIACH_CHANNEL_INSTANCE_ATTRS', 'attr': 'MQIACH_CHANNEL_STATUS', 'owner': 'Channel', 'metric': 'absolute', 'splitName': 'Channel', 'splitBy': 'MQCA_Q_NAME'}
        metrics['channelMessages'] = {'name': 'Channel.Messages', 'json_field': 'messages', 'paramIdentifier' : 'MQIACH_CHANNEL_INSTANCE_ATTRS', 'attr': 'MQIACH_MSGS', 'owner': 'Channel', 'metric': 'relative', 'splitName': 'Channel', 'splitBy': 'MQCACH_CHANNEL_NAME', 'aggregationField': 'totalMsgs'}
        metrics['channelBytesSent'] = {'name': 'Channel.BytesSent', 'json_field': 'bytesSent', 'paramIdentifier' : 'MQIACH_CHANNEL_INSTANCE_ATTRS', 'attr': 'MQIACH_BYTES_SENT', 'owner': 'Channel', 'metric': 'relative', 'splitName': 'Channel', 'splitBy': 'MQCACH_CHANNEL_NAME', 'aggregationField': 'totalBytesSent'}
        metrics['channelBytesReceived'] = {'name': 'Channel.BytesReceived', 'json_field': 'bytesReceived', 'paramIdentifier' : 'MQIACH_CHANNEL_INSTANCE_ATTRS', 'attr': 'MQIACH_BYTES_RECEIVED', 'owner': 'Channel', 'metric': 'relative', 'splitName': 'Channel', 'splitBy': 'MQCACH_CHANNEL_NAME', 'aggregationField': 'totalBytesReceived'}
        metrics['channelBuffersSent'] = {'name': 'Channel.BuffersSent', 'json_field': 'buffersSent', 'paramIdentifier' : 'MQIACH_CHANNEL_INSTANCE_ATTRS', 'attr': 'MQIACH_BUFFERS_SENT', 'owner': 'Channel', 'metric': 'relative', 'splitName': 'Channel', 'splitBy': 'MQCACH_CHANNEL_NAME', 'aggregationField': 'totalBuffersSent'}
        metrics['channelBuffersReceived'] = {'name': 'Channel.BuffersReceived', 'json_field': 'buffersReceived', 'paramIdentifier' : 'MQIACH_CHANNEL_INSTANCE_ATTRS', 'attr': 'MQIACH_BUFFERS_RECEIVED', 'owner': 'Channel', 'metric': 'relative', 'splitName': 'Channel', 'splitBy': 'MQCACH_CHANNEL_NAME', 'aggregationField': 'totalBuffersReceived'}
        metrics['channelLastMessageDate'] = {'name': 'Channel.LastMsgDate', 'json_field': 'lastMessageDate', 'paramIdentifier' : 'MQIACH_CHANNEL_INSTANCE_ATTRS', 'attr': 'MQCACH_LAST_MSG_DATE', 'owner': 'Channel', 'metric': 'absolute', 'splitName': 'Channel', 'splitBy': 'MQCACH_CHANNEL_NAME'}
        metrics['channelLastMessageTime'] = {'name': 'Channel.LastMsgTime', 'json_field': 'lastMessageTime', 'paramIdentifier' : 'MQIACH_CHANNEL_INSTANCE_ATTRS', 'attr': 'MQCACH_LAST_MSG_TIME', 'owner': 'Channel', 'metric': 'absolute', 'splitName': 'Channel', 'splitBy': 'MQCACH_CHANNEL_NAME'}
        metrics['channelLastMsg'] = {'name': 'Channel.LastMsg', 'paramIdentifier' : 'MQIACH_CHANNEL_INSTANCE_ATTRS', 'attr': 'MQCACH_LAST_MSG_TIME', 'owner': 'Channel', 'metric': 'relative', 'splitName': 'Channel', 'splitBy': 'MQCACH_CHANNEL_NAME'}
        metrics['channelCurrentSharingConvs'] = {'name': 'Channel.CurrentSharingConvs', 'json_field': 'currentSharingConvs', 'paramIdentifier' : 'MQIACH_CHANNEL_INSTANCE_ATTRS', 'attr': 'MQIACH_CURRENT_SHARING_CONVS', 'owner': 'Channel', 'metric': 'absolute', 'splitName': 'Channel', 'splitBy': 'MQCACH_CHANNEL_NAME', 'aggregationField': 'totalCurrentSharingConvs'}
        
        metrics['listenerStatus'] = {'name': 'Listener.ListenerStatus', 'json_field': 'status', 'paramIdentifier' : 'MQIACF_LISTENER_STATUS_ATTRS', 'attr': 'MQIACH_LISTENER_STATUS', 'owner': 'ChannelListener', 'metric': 'absolute', 'splitName': 'Listener', 'splitBy': 'MQCACH_LISTENER_NAME', 'normalValue': 'RUNNING'}
        
        
        config_params = {
            "port": hostPort, 
            "username": config['username'], 
            "password": "******", 
            "server_channel": config['server_connection'], 
            "queue_manager": self.queue_manager, 
            "endpoint": endpoint_id, 
            "tempdir": self.tempdir,
            "reset_q_stats": 'true' if runResetStats else 'false',
            "exclude_system": 'true' if config['exclude_system'] else 'false',
            "channels": ','.join(channels),
            "queues": ','.join(queues),
            #"listeners": ','.join(listeners),
            "model_q": config['model_queue'],
            "reply_q_prefix": config['reply_queue_prefix'],
            "dlq_alert": str(dlq_alert).lower(),
            "retry_channel_alert": str(retry_channel_alert).lower(),
            "get_topology": 'true' if 'get_queue_topology' in config and str(config['get_queue_topology']).lower() == 'true' else 'false',
            "use_ssl": 'true' if 'use_ssl' in config and str(config['use_ssl']).lower() == 'true' else 'false',
            "ssl_keystore": config['ssl_keystore'],
            "ssl_cipher": config['ssl_cipher'],
            "ssl_keystore_pswd": '******',
            "pluginVersion": kwargs["json_config"]["version"]}
        
        # Java does not have connection list (JMS has the class for it but Java does not) so we have to weirdly configure hostname and ports to accomodate for this
        # The last host in the connection list must not have a port included in the connection list and the port property must be set to the port of this last host
        # Explained here: https://www.capitalware.com/rl_blog/?p=5924
        if len(hostList) > 1:
            config_params['hostname'] = ','.join(hostList)
        else:
            config_params['hostname'] = hostIP

        logger.info(config_params)
        config_params["password"] = str(base64.b64encode(config['password'].encode('utf-8')), "utf-8")
        config_params["ssl_keystore_pswd"] = str(base64.b64encode(config['ssl_keystore_password'].encode('utf-8')), "utf-8")
        
        # Add the md5checksum
        config_params["md5checksum"] = hashlib.md5(str(config_params).encode()).hexdigest()
        
        confFile = self.tempdir +'/'+ endpoint_id + ".config"
        if os.path.isfile(confFile):
            logger.info("Config file exists, deleting it")
            try:
                os.unlink(confFile)
            except FileNotFoundError as e:
                logger.warn(f'Conf file "{confFile}" not found')
            except PermissionError as e:
                logger.error(f'Unable to delete conf file "{confFile}". Permission error: {e}')

        logger.info("Write data to " + confFile)
        with open(confFile, "w") as out_file:
            json.dump(config_params, out_file)
            
        logger.info(f'endpoint_id = {endpoint_id}')
        java_process = get_java_process([endpoint_id, "dynatrace_ibm_mq_connector.jar"])
        
        data_file = f'{self.tempdir}/{endpoint_id}.json'
        
        if java_process is None:
            logger.info("Attempting to start the Java Connector")
            java_process = start_java_process(endpoint_id,self.tempdir, config.get('java_path', ''))
            atexit.register(stop_java_process, java_process)
            get_java_process(["dynatrace_ibm_mq_connector.jar"])
        else:
            logger.info(f"Java Connector already started: {java_process}")
            # Check to see if data file is over 3 minutes old
            if os.path.isfile(data_file) and time.time()-os.path.getmtime(data_file) > 3*60:
                # Something happened that Java is not updating anymore the data file. Kill it
                logger.warn('Last time data was retrieved is over 3 minutes. Java might be hanging, stopping Java process')
                stop_java_process(java_process)
                try:
                    os.unlink(data_file)
                    os.unlink(f'{self.tempdir}/{endpoint_id}.lock')
                except FileNotFoundError as e:
                    logger.warn(f'Lock file or data file for {endpoint_id} not found in {self.tempdir}')
                except PermissionError as e:
                    logger.error(f'Unable to delete lock file or data file for {endpoint_id} in {self.tempdir}. Permission error: {e}')
                finally:
                    return
            
        logger.info(f'Looking for {data_file}')
        if os.path.isfile(data_file):
        
            # If we get to this point. Data file is current
            with open(data_file) as json_file:
                logger.debug(f'Opened file {data_file}')
                data = json.load(json_file)
                
                # If an error occurred, display it
                if data['error']['error'] == True and 'MQRC_UNKNOWN_OBJECT_NAME' not in data["error"]["errorMessage"]:
                    logger.error(f'Unable to connect to queue manager {self.queue_manager}. {data["error"]["errorMessage"]}')
                    if 'MQRC_HOST_NOT_AVAILABLE' in data["error"]["errorMessage"]:
                        self.setMetric('absolute', 'QueueManager.Status', 0, {})
                    else:
                        props = {'Host': hostIP, 'Port': hostPort, 'Queue manager': self.queue_manager, 'Server-connection': config['server_connection']}
                        self.node.report_error_event(f'Unable to get data from queue manager {self.queue_manager}. {data["error"]["errorMessage"]}', 'Could not get data from queue manager')
                else:    
                
                    data_props = data['metrics']['properties']
                    version = f'{int(data_props["version"][0:2])}.{int(data_props["version"][2:4])}.{int(data_props["version"][4:6])}' if data_props['version'] != '' else "Unknown version"
                    if data_props.get("platformCode") in self.OS:
                        OSname = f'IBM MQ {version} on {self.OS[data_props.get("platformCode")]}'
                    else:
                        OSname = f'Unknown[{data_props.get("platformCode")}]'
                        
                    logger.debug(f'Version = {OSname}')
                    self.node.report_property('Version', OSname)
                    self.node.report_property('Max Priority', str(data_props.get('maxPriority')))
                    
                    # Set queue manager metrics
                    if 'queueManager' in data['metrics']:
                        logger.debug('Reporting Queue manager metrics')
                        for each_metric in data['metrics']['queueManager']:
                            #logger.info(f'each queue manager metric: {each_metric}')
                            val = data['metrics']['queueManager'][each_metric]
                            if each_metric in metrics:
                                if 'normalValue' in metrics[each_metric]:
                                    val = 100 if val == metrics[each_metric]['normalValue'] else 0
                                self.setMetric(metrics[each_metric]['metric'], metrics[each_metric]['name'], float(val), {})
                    
                    # Alert on DLQ > 0
                    if dlq_alert == 'true' and data_props.get('deadLetterQueueName') != '' and int(data_props.get('deadLetterQueueDepth')) > 0:
                        logger.debug(f"DLQ alert: Sending error event. Depth = {data_props.get('deadLetterQueueDepth')}")
                        self.node.report_error_event(f'Dead letter queue {data_props.get("deadLetterQueueName")} on {self.queue_manager} has {data_props.get("deadLetterQueueDepth")} undeliverable message(s)', f'DLQ {data_props.get("deadLetterQueueName")} on {self.queue_manager} has message(s)')
                            
                    if 'queues' in data['metrics'] and len(queues) > 0:
                        if len(data['metrics']['queues']) > 0:
                            logger.info(f"Reporting queue metrics. Results = {len(data['metrics']['queues'])}")
                            #data['metrics']['queues'] = self.filterList(data['metrics']['queues'], queues, config['exclude_system'], 'name')
                            #logger.info(f"data = {data['metrics']['queues']}")
                            for i in range(len(data['metrics']['queues'])):
                                lastGetDate,lastGetTime,lastPutDate,lastPutTime = None,None,None,None
                                queue = data['metrics']['queues'][i]
                                for each in queue:
                                    #logger.info(f'data["metrics"]["queues"][i] = {data["metrics"]["queues"][i]}')
                                    #logger.info(f'each = {each}')
                                    if isinstance(queue[each], dict):
                                        for each_metric in queue[each]:
                                            #logger.debug(f'each queue metric: {each_metric}')
                                            if each_metric in metrics:
                                                val = queue[each][each_metric]
                                                if 'normalValue' in metrics[each_metric]:
                                                    val = 100 if val == metrics[each_metric]['normalValue'] else 0
                                                if each_metric == 'queueLastGetDate' and val != '':
                                                    lastGetDate = datetime.strptime(val, '%Y-%m-%d')
                                                elif each_metric == 'queueLastGetTime' and val != '':
                                                    lastGetTime = datetime.strptime(val, '%H.%M.%S')
                                                elif each_metric == 'queueLastPutDate' and val != '':
                                                    lastPutDate = datetime.strptime(val, '%Y-%m-%d')
                                                elif each_metric == 'queueLastPutTime' and val != '':
                                                    lastPutTime = datetime.strptime(val, '%H.%M.%S')
                                                elif each_metric == 'queueMaxQueueDepth' and val != '':
                                                    currentQueueDepth = int(queue['status']['queueCurrentQueueDepth'])
                                                    percentQueueDepth = (currentQueueDepth * 100) / int(val) if int(val) > 0 else 0
                                                    self.setMetric(metrics[each_metric]['metric'], 'Queue.PercentQueueDepth', percentQueueDepth, {'Queue': queue['name']})
                                                else:
                                                    if val != '' and val != -1:
                                                        self.setMetric(metrics[each_metric]['metric'], metrics[each_metric]['name'], float(val), {'Queue': queue['name']})
                                                        #logger.info(f'Set metric {metrics[each_metric]["name"]} with val {val} for queue {queue["name"]}')
                                if lastGetDate and lastGetTime:
                                    lastGetDateTimeDelta = datetime.now()-datetime.combine(lastGetDate.date(), lastGetTime.time())
                                    #logger.debug(f'lastGetDateTimeDelta: {lastGetDateTimeDelta}')
                                    lastGetDateTime = (lastGetDateTimeDelta.seconds*1000)+(lastGetDateTimeDelta.microseconds/1000)
                                    self.setMetric('relative', 'Queue.LastGet', lastGetDateTime, {'Queue': queue['name']})
                                
                                if lastPutDate and lastPutTime:
                                    lastPutDateTimeDelta = datetime.now()-datetime.combine(lastPutDate.date(), lastPutTime.time())
                                    #logger.debug(f'lastPutDateTimeDelta: {lastPutDateTimeDelta}')
                                    lastPutDateTime = (lastPutDateTimeDelta.seconds*1000)+(lastPutDateTimeDelta.microseconds/1000)
                                    self.setMetric('relative', 'Queue.LastPut', lastPutDateTime, {'Queue': queue['name']})
                                    
                                if 'reset' in queue and queue['reset']['queueTimeSinceLastReset'] >= 0:
                                    time_since_last_reset = queue['reset']['queueTimeSinceLastReset']
                                    # Calculate enqueue/dequeue rates
                                    self.setMetric(metrics['queueEnqueue']['metric'], 'Queue.EnqueueRate', float(queue['reset']['queueEnqueue']/time_since_last_reset) if time_since_last_reset > 0 else 0, {'Queue': queue['name']})
                                    self.setMetric(metrics['queueDequeue']['metric'], 'Queue.DequeueRate', float(queue['reset']['queueDequeue']/time_since_last_reset) if time_since_last_reset > 0 else 0, {'Queue': queue['name']})
                                    
                    if 'channels' in data['metrics'] and len(channels) > 0:
                        if len(data['metrics']['channels']) > 0:
                            logger.debug(f"Reporting channel metrics. Results = {len(data['metrics']['channels'])}")
                            #data['metrics']['channels'] = self.filterList(data['metrics']['channels'], channels, config['exclude_system'], 'channelName')
                            for each_channel in data['metrics']['channels']:
                                if each_channel['channelName'] not in channel_totals:
                                    channel_totals[each_channel['channelName']] = {}
                                    
                                lastMsgDate, lastMsgTime = None, None
                                for each_metric in each_channel:
                                    #logger.debug(f'each channel metric: {each_metric}')
                                    if each_metric in metrics:
                                        val = each_channel[each_metric]
                                        if each_metric == 'channelStatus':
                                            if val == 'MQCHS_RETRYING' and retry_channel_alert:
                                                self.node.report_error_event(f'Channel {each_channel["channelName"]} on {self.queue_manager} is in Retrying state', f'Channel {each_channel["channelName"]} in Retrying state', properties={'Queue manager': self.queue_manager, 'Channel': each_channel["channelName"], 'Host': ','.join(hostList)})
                                            self.node.state_metric('Channel.Status', self.channel_states[val], {'Channel': each_channel['channelName']})
                                            if val == 'INACTIVE':
                                                break
                                        elif 'normalValue' in metrics[each_metric]:
                                            val = 100 if val == metrics[each_metric]['normalValue'] else 0
                                            self.setMetric(metrics[each_metric]['metric'], metrics[each_metric]['name'], float(val), {'Channel': each_channel['channelName']})
                                        elif each_metric == 'channelLastMessageDate':
                                            lastMsgDate = datetime.strptime(val, '%Y-%m-%d') if val != '' else None
                                        elif each_metric == 'channelLastMessageTime':
                                            lastMsgTime = datetime.strptime(val, '%H.%M.%S') if val != '' else None
                                        else:
                                            if val != '':
                                                if each_metric not in channel_totals[each_channel['channelName']]:
                                                    channel_totals[each_channel['channelName']][each_metric] = val
                                                else:
                                                    #logger.debug(f"Adding to {each_metric} total for {each_channel['channelName']} = {channel_totals[each_channel['channelName']][each_metric]}+{val}")
                                                    channel_totals[each_channel['channelName']][each_metric] = channel_totals[each_channel['channelName']][each_metric]+val
                                        
                                if isinstance(lastMsgDate, datetime) and isinstance(lastMsgTime, datetime):
                                    lastMsgDateTime = datetime.combine(lastMsgDate.date(), lastMsgTime.time())
                                    lastMsgDateTime = (datetime.now()-lastMsgDateTime).total_seconds()
                                    
                                    if ('channelLastMsg' not in channel_totals[each_channel['channelName']]) or ('channelLastMsg' in channel_totals[each_channel['channelName']] and channel_totals[each_channel['channelName']]['channelLastMsg'] > lastMsgDateTime):
                                        logger.debug(f"lastMsgDateTime for {each_channel['channelName']} = {lastMsgDateTime}")
                                        channel_totals[each_channel['channelName']]['channelLastMsg'] = lastMsgDateTime

                    i = 0
                    for each_channel in channel_totals:
                        if i >= MAX_ITEMS:
                            break
                        for each_metric in channel_totals[each_channel]:
                            #logger.debug(f'Final {each_metric}: {channel_totals[each_channel][each_metric]} for channel {each_channel}')
                            self.setMetric(metrics[each_metric]['metric'], metrics[each_metric]['name'], float(channel_totals[each_channel][each_metric]), {'Channel': each_channel})
                        i = i+1

                    if 'listeners' in data['metrics'] and len(listeners) > 0:
                        logger.info('Reporting listener metrics')
                        if len(data['metrics']['listeners']) > 0:
                            data['metrics']['listener'] = self.filterList(data['metrics']['listeners'], listeners, config['exclude_system'], 'listenerName')
                            for each_listener in data['metrics']['listeners']:
                                val = each_listener['listenerStatus']
                                val = 100 if val == metrics['listenerStatus']['normalValue'] else 0
                                self.setMetric(metrics['listenerStatus']['metric'], metrics['listenerStatus']['name'], float(val), {'Listener': f"{each_listener['listenerName']}.{each_listener['listenerProtocol']}:{each_listener['listenerPort']}"})
                    
                    if config['get_queue_topology'] and (self.last_topology_retrieve[self.queue_manager]['time'] == 0 or time.time()-self.last_topology_retrieve[self.queue_manager]['time'] > 3600):
                        try:
                            if data['clusterData']['name'] != '':
                                cluster_data = json.dumps(data['clusterData'], cls=EncodeText)
                                self.send_topology(cluster_data, api_token, tenant)
                        except Exception as e:
                            logger.exception(f'Could not send {self.queue_manager} topology. Reason: {e}')
                            self.report_custom_info_event(f'Could not send {self.queue_manager} topology.', 'Could not send queue manager topology')
                    
        logger.info("Finished polling")
    # Set metric according to metric type: absolute or relative
    def setMetric(self, metricType, keyName, val, metricDimension):
        if (metricType == 'absolute'):
            self.node.absolute(key=keyName, value=val, dimensions=metricDimension)
        else:
            key = self.queue_manager+'_'+list(metricDimension.values())[0]+'_'+keyName
            if key not in self.metricHolder:
                self.metricHolder[key] = val
            else:
                deltaVal = abs(self.metricHolder[key]-val)
                self.metricHolder[key] = val
                #logger.debug(f'Setting Delta: {deltaVal} for dimension: {metricDimension}')
                self.node.absolute(key=keyName, value=deltaVal, dimensions=metricDimension)
    
    def filterList(self, data, search_patterns, excludeSystem, field):
        results, exclusions, inclusions = [],[],[]
        for pattern in search_patterns:
            if pattern.startswith('-') or pattern.startswith('~'):
                exclusions.append(pattern[1:])
            elif '*' not in pattern:
                try:
                    results.append(next((item for item in data if item[field] == pattern)))
                except StopIteration:
                    logger.info(f'Could not find {pattern} in data set')
            else:
                logger.debug(f'Added pattern to inclusion: {pattern}')
                inclusions.append(pattern)

        # Filter out any SYSTEM results if selected
        if excludeSystem == True:
            exclusions.append('SYSTEM.*')
        
        # Run filters on the results and join remaining results
        itemsExcluded = list(self.search_for_items(data, exclusions, field))
        itemsIncluded = list(self.search_for_items(data, inclusions, field))
        
        results = results + [i for i in itemsIncluded if i not in itemsExcluded]
        
        # Remove empties
        results = [i for i in results if i]
        
        #logger.info(f'results = {results}')
        # Still limit the results to MAX_ITEMS
        return results[0:MAX_ITEMS]
        
    def search_for_items(self, items, patterns, field):
        for item in items:
            if any(fnmatch.fnmatch(item[field], pattern) for pattern in patterns):
                yield item
        
    def send_topology(self,data,apiToken,tenant):
        try:
            if not tenant.endswith('/'):
                tenant += '/'
            if not tenant.startswith('http'):
                tenant = 'https://'+tenant
                
            url = f'{tenant}api/config/v1/service/ibmMQTracing/queueManager/{self.queue_manager}'
            
            # Put the data
            response = requests.put(f'{url}', proxies=self.request_proxy_list, data=data, headers={'Content-Type': 'application/json', 'Authorization': f'Api-Token {apiToken}'}, verify=False)
            
            if response.status_code == 204 or response.status_code == 201:
                logger.info('Succeeded sending topology')
                self.last_topology_retrieve[self.queue_manager]['time'] = time.time()
                return True
            else:
                logger.error(f'Error sending topology to API {url}. Response code: {response.status_code}. {response.text}. Data: {data}')
                self.node.report_error_event(f'An error occurred sending queue manager {self.queue_manager} topology mapping to environment', 'Error sending MQ topology', properties={'Queue manager': self.queue_manager, 'Error code': str(response.status_code), 'Response': response.text})
                return False
        except Exception as e:
            logger.error(f'Failed to communicate with tenant {url}. {e}')
            self.node.report_error_event(f'An error occurred sending queue manager {self.queue_manager} topology mapping to environment. Check plugin log for further error details.', 'Error sending MQ topology', properties={'Queue manager': self.queue_manager})
            return False
            
    def report_custom_info_event(self, message, title, properties={}):
        if self.events < MAX_EVENTS:
            if title + message not in self.previousEvents:
                self.events += 1
                self.node.report_custom_info_event(message, title, properties)
                self.previousEvents.append(title + message)